from . import left, matches, right, smap
from neer_match.reasoning import RefutationModel
import pytest
import tensorflow as tf


def test_refutation():
    """Test refutation model."""
    model = RefutationModel(smap)
    model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=0.001))
    model.fit(left, right, matches, refutation="title", epochs=1)
    assert model is not None, "Failed to fit refutation model with association."
    model.fit(left, right, matches, refutation={"title": "jaro_winkler"}, epochs=1)
    assert (
        model is not None
    ), "Failed to fit refutation model with association and similarity."
    model = RefutationModel(smap)
    model.compile()
    with pytest.raises(ValueError):
        model.fit(left, right, matches, refutation="no-title", epochs=1)
    with pytest.raises(ValueError):
        model.fit(left, right, matches, refutation={"title": "no-similarity"}, epochs=1)
    with pytest.raises(ValueError):
        model.fit(left, right, matches, refutation={"title": []}, epochs=1)
    with pytest.raises(ValueError):
        model.fit(left, right, matches, refutation={"title"}, epochs=1)



"""
Record pair network module.

This module contains functionality for instantiating, training, and using a record
 pair networks.
"""

from neer_match.field_pair_network import FieldPairNetwork
from neer_match.similarity_map import SimilarityMap
import tensorflow as tf
import typing


class RecordPairNetwork(tf.keras.Model):
    """Record network class.

    The class creates networks for matching records from two datasets. The networks
    consist of field pair networks, constructed according to the passed similarity map,
    that are concatenated and passed through a series of hidden dense layers. The output
    layer has a sigmoid activation function.

    The depth and width of the record pair hidden layers is specified by the
    initial_record_width_scale and record_depth parameters. The width of the initial
    hidden layers is calculated multiplying the initial_record_width_scale by the number
    of field pairs (i.e., the number of associations in the similarity map, see
    :meth:`.no_associations`). The widths of the subsequent hidden layers are
    calculated dividing the initial width by the layer depth-index plus one.

    The depth and width of the field pair networks are specified by the
    initial_feature_width_scales and  feature_depths parameters (see
    :class:`.FieldPairNetwork` for more details).

    Attributes:
        similarity_map (SimilarityMap): The similarity map object.
        initial_feature_width_scales (list[int]): The initial width scales of the hidden
            layers for each field pair network.
        feature_depths (list[int]): The depths of the networks for each field pair
            network.
        initial_record_width_scale (int): The initial width scale of the hidden layers
            for the record pair network.
        record_depth (int): The depth of the record pair network.
    """

    def __init__(
        self,
        similarity_map: SimilarityMap,
        initial_feature_width_scales: typing.Union[int, list[int]] = 10,
        feature_depths: typing.Union[int, list[int]] = 2,
        initial_record_width_scale: int = 10,
        record_depth: int = 4,
        **kwargs,
    ) -> None:
        """Initialize a record network object.

        Args:
            similarity_map: The similarity map.
            initial_feature_width_scales: The initial width scales of the hidden layers
                    for each field pair network. If an integer is passed, the same scale is
                    used for all networks.
            feature_depths: The depths of the networks for each field pair network. If
                    an integer is passed, the same depth is used for all networks.
            initial_record_width_scale: The initial width scale of the hidden layers for
                    the record pair network.
            record_depth: The depth of the record pair network.
            **kwargs: Additional keyword arguments passed to parent class
                      (:class:`tensorflow.keras.Model`).
        """
        if not isinstance(similarity_map, SimilarityMap):
            raise ValueError(
                "Input similarity_map must be an instance of SimilarityMap."
            )
        if (
            not isinstance(initial_record_width_scale, int)
            or initial_record_width_scale < 1
        ):
            raise ValueError(
                "Input initial_record_width_scale must be a positive integer."
            )
        if not isinstance(record_depth, int) or record_depth < 1:
            raise ValueError("Input record_depth must be a positive integer.")
        # The remaining arguments are check in FieldPairNetwork.

        self.similarity_map = similarity_map
        self.initial_feature_width_scales = initial_feature_width_scales
        self.feature_depths = feature_depths
        self.initial_record_width_scale = initial_record_width_scale
        self.record_depth = record_depth

        no_assoc = similarity_map.no_associations()
        if isinstance(initial_feature_width_scales, int):
            initial_feature_width_scales = [initial_feature_width_scales] * no_assoc
        if isinstance(feature_depths, int):
            feature_depths = [feature_depths] * no_assoc

        super().__init__(**kwargs)

        self.field_networks = []
        for i, name in enumerate(similarity_map.association_names()):
            self.field_networks.append(
                FieldPairNetwork(
                    size=similarity_map.association_sizes()[i],
                    initial_width_scale=initial_feature_width_scales[i],
                    depth=feature_depths[i],
                    name=name,
                )
            )
        self.concat = tf.keras.layers.Concatenate()
        self.record_layers = []
        for i in range(record_depth):
            size = (initial_record_width_scale * no_assoc) // (i + 1)
            self.record_layers += [
                tf.keras.layers.Dense(
                    max(size, 2),
                    activation=tf.keras.activations.relu,
                    name=f"hidden_record_mixing_{i}",
                )
            ]
        self.record_layers += [
            tf.keras.layers.Dense(
                1,
                tf.keras.activations.sigmoid,
                name="record_classifier",
            )
        ]

    def get_config(self) -> dict:
        """Return the configuration of the network."""
        config = super().get_config().copy()
        config.update(
            {
                "similarity_map": self.similarity_map,
                "initial_feature_width_scales": self.initial_feature_width_scales,
                "feature_depths": self.feature_depths,
                "initial_record_width_scale": self.initial_record_width_scale,
                "record_depth": self.record_depth,
            }
        )
        return config

    def build(
        self,
        input_shapes: typing.Union[
            typing.List[tf.TensorShape], typing.Dict[str, tf.TensorShape]
        ],
    ) -> None:
        """Build the network."""
        field_output_shapes = []
        if isinstance(input_shapes, dict):
            input_shapes = list(input_shapes.values())

        for i, input_shape in enumerate(input_shapes):
            self.field_networks[i].build(input_shape)
            field_output_shapes.append((input_shape[0], 1))
        self.concat.build(field_output_shapes)
        input_shapes = self.concat.compute_output_shape(field_output_shapes)
        for layer in self.record_layers:
            layer.build(input_shapes)
            input_shapes = (input_shapes[0], layer.units)
        super().build(input_shapes)

    def call(
        self, inputs: typing.Union[typing.List[tf.Tensor], typing.Dict[str, tf.Tensor]]
    ) -> tf.Tensor:
        """Run the network on input."""
        if isinstance(inputs, dict):
            inputs = list(inputs.values())
        outputs = []
        for i, x in enumerate(inputs):
            outputs.append(self.field_networks[i](x))
        output = self.concat(outputs)
        for layer in self.record_layers:
            output = layer(output)
        return output



"""
Global explainability module.

This module provides customized global explainability functionality for entity matching
tasks.
"""

from neer_match.data_generator import DataGenerator
from neer_match.matching_model import (
    _matching_model_or_raise,
    DLMatchingModel,
    NSMatchingModel,
)
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
import typing


def partial_dependence_function(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    xfeatures: dict[str, float],
) -> float:
    """Calculate the partial dependence of the model on the given keys.

    Replaces the values of the given keys with the specified values and
    calculates the average prediction of the model.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        xfeatures: A dictionary of the features with values where the partial
            dependence is calculated.
    """
    _matching_model_or_raise(model)
    if not isinstance(xfeatures, dict):
        raise ValueError("The input xfeatures must be a dictionary")
    for k, v in xfeatures.items():
        if not isinstance(k, str):
            raise ValueError("The keys of xfeatures must be strings")
        if k not in model.similarity_map.keys():
            raise ValueError(f"Key {k} is not in the similarity map")
        if not isinstance(v, float) or v < 0 or v > 1:
            raise ValueError(
                "The values of xfeatures must be floats in the range [0, 1]"
            )

    generator = DataGenerator(
        model.record_pair_network.similarity_map,
        left,
        right,
    )

    offsets = model.similarity_map.association_offsets()
    names = model.similarity_map.association_names()

    xkeys = list(xfeatures.keys())
    positions = [model.similarity_map.keys().index(key) for key in xkeys]
    positions = {
        (y := len([x for x in offsets if x <= pos]) - 1): (pos - offsets[y])
        for pos in positions
    }

    xvalues = list(xfeatures.values())
    result = 0.0
    for features in generator:
        for i, (apos, spos) in enumerate(positions.items()):
            features[names[apos]][:, spos] = xvalues[i]
        preds = model.record_pair_network(features)
        result += tf.reduce_sum(preds)
    result /= generator.no_pairs()
    return result.numpy()


def partial_dependence(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    key: str,
    n: int = 50,
) -> np.ndarray:
    """Calculate the partial dependence of a key over a domain grid.

    Creates a [0, 1] grid of n interpolation points and calculates the partial
    dependence of the model on the key at each point.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        key: The key for which to calculate the partial dependence.
        n: The number of interpolation points to use.
    """
    if not isinstance(n, int) or n < 2:
        raise ValueError("Interpolation points (n) must be an integer greater than 1")
    # The remaining arguments are validated in partial_dependence_function
    return np.array(
        [
            partial_dependence_function(model, left, right, {key: i / (n - 1)})
            for i in range(n)
        ]
    )


def partial_dependence_feature_importance(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    key: str,
    n: int = 50,
):
    """Calculate the feature importance of a key using partial dependence.

    Calculates the standard deviation of the partial dependence of the model on the key
    over a [0, 1] domain grid.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        key: The key for which to calculate the feature importance.
        n: The number of interpolation points to use.
    """
    # Input arguments are validated in partial_dependence
    return np.std(partial_dependence(model, left, right, key, n))


def partial_dependence_plot(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    key: str,
    n: int = 50,
):
    """Plot the partial dependence of a key.

    Plots the partial dependence of the model on the key over a [0, 1] domain grid.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        key: The key for which to calculate the partial dependence.
        n: The number of interpolation points
    """
    domain = [i / (n - 1) for i in range(n)]
    values = partial_dependence(model, left, right, key, n)
    std = np.std(values)
    values_up = values + std
    values_down = values - std
    fig, ax = plt.subplots()
    ax.plot(domain, values, label="PD", color="blue", linestyle="-")
    ax.plot(domain, values_up, label="PD $\\pm$ std", color="blue", linestyle="--")
    ax.plot(domain, values_down, color="blue", linestyle="--")
    ax.legend()
    ax.set_xlabel(key)
    ax.set_ylabel("Probability")
    return fig


def accumulated_local_effect(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    xkey: str,
    xvalue: float,
    centered: bool = True,
    k: int = 50,
):
    """Calculate the accumulated local effect of a key over a domain grid.

    Creates a 0 to xvalue grid of n interpolation points, calculates local
    differences in the model's predictions for each segment of the grid, and
    averages the differences. If centered is True, the partial dependence of the
    model on the key at xvalue is subtracted from the accumulated local effect.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        xkey: The key for which to calculate the accumulated local effect.
        xvalue: The value of at which the accumulated local effect is calculated.
        centered: Whether to center the accumulated local effect.
        k: The number of interpolation points
    """
    _matching_model_or_raise(model)
    if not isinstance(xkey, str):
        raise ValueError("The xkey argument must be a string")
    if xkey not in model.similarity_map.keys():
        raise ValueError(f"Key {xkey} is not in the similarity map")
    if xvalue < 0 or xvalue > 1:
        raise ValueError("xvalue must be in the range [0, 1]")
    if not isinstance(centered, bool):
        raise ValueError("The centered argument must be a boolean")
    if not isinstance(k, int) or k < 1:
        raise ValueError("The k argument must be an integer greater than 0")
    # The dataset arguments are validated in partial_dependence_function

    generator = DataGenerator(
        model.record_pair_network.similarity_map,
        left,
        right,
    )

    offsets = model.similarity_map.association_offsets()
    names = model.similarity_map.association_names()

    pos = model.similarity_map.keys().index(xkey)
    apos = len([x for x in offsets if x <= pos]) - 1
    spos = pos - offsets[apos]

    result = 0.0
    for i in range(k):
        term = 0.0
        count = 0
        for features in generator:
            x = features[names[apos]][:, spos]
            lvalue = xvalue * i / (k + 1)
            hvalue = xvalue * (i + 1) / (k + 1)
            xi = (x >= lvalue) & (x <= hvalue)
            used = {k: f[xi, :] for k, f in features.items()}
            used[names[apos]][:, spos] = lvalue
            lpreds = model.record_pair_network(used)
            used[names[apos]][:, spos] = hvalue
            hpreds = model.record_pair_network(used)
            term += tf.reduce_sum(hpreds - lpreds).numpy()
            count += sum(x)
        result += term / count
    if centered:
        result -= partial_dependence_function(model, left, right, {xkey: xvalue})
    return result


def accumulated_local_effect_plot(
    model: typing.Union[DLMatchingModel, NSMatchingModel],
    left: pd.DataFrame,
    right: pd.DataFrame,
    key: str,
    centered: bool = True,
    n: int = 50,
    k: int = 50,
):
    """Plot the accumulated local effect of a key.

    Args:
        model: The matching model to use.
        left: The left DataFrame.
        right: The right DataFrame.
        key: The key for which to calculate the accumulated local effect.
        centered: Whether to center the accumulated local effect.
        n: The number of interpolation points for the figure.
        k: The number of interpolation points for the local effect.
    """
    if not isinstance(n, int) or n < 2:
        raise ValueError("The n argument must be an integer greater than 1")
    # The remaining arguments are validated in accumulated_local_effect
    domain = [i / n for i in range(n + 1)]
    values = [accumulated_local_effect(model, left, right, key, i, k=k) for i in domain]
    std = np.std(values)
    values_up = values + std
    values_down = values - std
    fig, ax = plt.subplots()
    ax.plot(domain, values, label="ALE", color="blue", linestyle="-")
    ax.plot(domain, values_up, label="ALE $\\pm$ std", color="blue", linestyle="--")
    ax.plot(domain, values_down, color="blue", linestyle="--")
    ax.legend()
    ax.set_xlabel(key)
    ax.set_ylabel("$\\Delta$ Probability")
    return fig



"""
Similarity encoding module.

The module provides functionality to store and manage a similarity encoders.
"""

from neer_match.similarity_map import available_similarities, SimilarityMap
import numpy
import pandas
import typing


class SimilarityEncoder:
    """Similarity encoder class.

    The class creates a similarity encoder from a similarity map. It can be used to
    encode pairs of records from two datasets.

    Attributes:
        similarity_map (SimilarityMap): The similarity map object.
        scalls (list[str]): The similarity function names.
        no_scalls (int): The number of similarity calls (field pairs and similarities).
        no_assoc (int): The number of associations (field pairs).
        assoc_begin (numpy.ndarray): The beginning offsets of the associations.
        assoc_sizes (numpy.ndarray): The sizes (number of used similarities) of the
            associations.
        assoc_end (numpy.ndarray): The ending indices of the associations.
    """

    def __init__(self, similarity_map: SimilarityMap) -> None:
        """Initialize a similarity encoder object.

        Args:
            similarity_map: The similarity map.
        """
        if not isinstance(similarity_map, SimilarityMap):
            raise ValueError(
                "Input similarity_map must be an instance of SimilarityMap. "
                f"Instead got {type(similarity_map)}"
            )
        self.similarity_map = similarity_map
        self.scalls = []

        for i in range(len(self.similarity_map)):
            lcol, rcol, sim = self.similarity_map[i]
            scall = available_similarities()[sim]
            if scall is None:
                raise ValueError(f"Unknown similarity function: {sim}")
            self.scalls.append(scall)
        self.no_scalls = len(self.scalls)
        self.no_assoc = self.similarity_map.no_associations()
        self.assoc_begin = numpy.array(self.similarity_map.association_offsets())
        self.assoc_sizes = numpy.array(self.similarity_map.association_sizes())
        self.assoc_end = self.assoc_begin + self.assoc_sizes

    def __call__(
        self, left: pandas.DataFrame, right: pandas.DataFrame
    ) -> typing.List[numpy.ndarray]:
        """Encode one or more pair of records.

        Calculate the similarities for each association (field pair) and return them
        in a list of arrays.

        Args:
            left: The left dataset.
            right: The right dataset.
        """
        sim_matrix = self.encode_as_matrix(left, right)
        return [
            sim_matrix[:, self.assoc_begin[i] : self.assoc_end[i]]
            for i in range(self.no_assoc)
        ]

    def encode_as_matrix(
        self, left: pandas.DataFrame, right: pandas.DataFrame
    ) -> numpy.ndarray:
        """Encode a pair of records as a matrix.

        Calculate the similarities for each association (field pair) and return them
        all stacked together in a matrix (i.e., the similarity matrix).

        Args:
            left: The left dataset.
            right: The right dataset.
        """
        if left.shape[0] != right.shape[0]:
            raise ValueError(
                f"Left and right datasets must have the same number of records. "
                f"Instead got {left.shape[0]} (left) and {right.shape[0]} (right)."
            )

        lx = left[self.similarity_map.lcols]
        rx = right[self.similarity_map.rcols]

        if len(lx.shape) == 1:
            vector = numpy.array(
                [self.scalls[i](lx.iloc[i], rx.iloc[i]) for i in range(self.no_scalls)]
            )
            vector.shape = (1, vector.shape[0])
            return vector
        else:
            return numpy.array(
                [
                    [
                        self.scalls[j](lx.iloc[i, j], rx.iloc[i, j])
                        for j in range(self.no_scalls)
                    ]
                    for i in range(lx.shape[0])
                ]
            )

    def encoded_shape(
        self, batch_size: int = -1
    ) -> typing.List[typing.Tuple[int, int]]:
        """Return the shape of the encoded data."""
        return [(batch_size, sz) for sz in self.assoc_sizes]

    def report_encoding(
        self,
        left: typing.Union[pandas.Series, pandas.DataFrame],
        right: typing.Union[pandas.Series, pandas.DataFrame],
    ) -> typing.List[pandas.DataFrame]:
        """Report encoding of a pair of records.

        Calculate the similarities for each association (field pair) in the similarity
        map and return them in a list of data frames. The function expects that the
        left and right datasets have the same number of records. It does not operate on
        the cross product of the records, but rather on the records at the same position
        in both datasets.

        Args:
            left: The left dataset.
            right: The right dataset.
        """
        smatrix = self.encode_as_matrix(left, right)
        report = []
        if isinstance(left, pandas.Series) and isinstance(right, pandas.Series):
            to = 1
        elif isinstance(left, pandas.DataFrame) and isinstance(right, pandas.DataFrame):
            to = left.shape[0]
        else:
            raise ValueError("Left and right must be a pandas Series or DataFrame.")
        for pos in range(to):
            if isinstance(left, pandas.Series):
                lseries = left[self.similarity_map.lcols]
                rseries = right[self.similarity_map.rcols]
            else:
                lseries = left.iloc[pos][self.similarity_map.lcols]
                rseries = right.iloc[pos][self.similarity_map.rcols]
            sseries = pandas.Series(smatrix[pos, :].tolist())
            lseries.name = "Left"
            lseries.index = self.similarity_map.keys()
            rseries.name = "Right"
            rseries.index = self.similarity_map.keys()
            sseries.name = "Similarities"
            sseries.index = self.similarity_map.keys()
            report.append(pandas.concat([lseries, rseries, sseries], axis=1))
        return report



"""
Entity matching data generator module.

This module provides a data generation functionality for entity matching tasks.
"""

from neer_match.similarity_map import SimilarityMap
from neer_match.similarity_encoding import SimilarityEncoder
import numpy as np
import pandas as pd
import tensorflow as tf
import typing


class DataGenerator(tf.keras.utils.Sequence):
    """Data generator class.

    The class provides a data generator for entity matching tasks. It inherits
    from the tf.keras.utils.Sequence class. Instances generate
    batches of similarities for the associated fields of two records in the
    cross product of the left and right data frames. The cross product is not explicitly
    computed. Instead, instances emulate it using indexing calculations.

    Attributes:
        left (pandas.DataFrame): The left DataFrame.
        right (pandas.DataFrame): The right DataFrame.
        matches (pandas.DataFrame): The matches DataFrame.
        batch_size (int): Batch size.
        mismatch_share (float): Mismatches share.
        shuffle (bool): Shuffle flag.
        full_size (int): The number of potential the record pairs.
        used_size (int): The used size of the record pairs.
        no_used_mismatches_per_match (int): The number of used mismatches per match.
        no_batches (int): The number of batches per epoch.
        last_batch_size (int): The size of the last batch.
        similarity_map (SimilarityMap): The similarity map.
        similarity_encoder (SimilarityEncoder): The similarity encoder.
        indices (numpy.ndarray): The used indices for the record pairs.
    """

    def __init__(
        self,
        similarity_map: SimilarityMap,
        left: pd.DataFrame,
        right: pd.DataFrame,
        matches: pd.DataFrame = None,
        batch_size: int = 32,
        mismatch_share: float = 0.1,
        shuffle: bool = False,
    ) -> None:
        """Initialize a data generator object.

        Prepare the indexing variables that are used in the data generation process.

        Args:
            similarity_map: A similarity map object.
            left: The left DataFrame.
            right: The right DataFrame.
            matches: The matches DataFrame.
            batch_size: Batch size.
            mismatch_share: Mismatches share.
            shuffle: Shuffle flag.
        """
        if not isinstance(left, pd.DataFrame):
            raise ValueError("Input left must be a pandas DataFrame.")
        if not isinstance(right, pd.DataFrame):
            raise ValueError("Input right must be a pandas DataFrame.")
        if not isinstance(matches, pd.DataFrame) and matches is not None:
            raise ValueError("Input matches must be a pandas DataFrame or None.")
        if not isinstance(batch_size, int) or batch_size < 1:
            raise ValueError("Input batch_size must be a positive integer.")
        if (
            not isinstance(mismatch_share, float)
            or mismatch_share < 0.0
            or mismatch_share > 1.0
        ):
            raise ValueError("Input mismatch_share must be a float in [0, 1].")
        if not isinstance(shuffle, bool):
            raise ValueError("Input shuffle must be a boolean.")
        # The similarity_map is check in SimilarityEncoder and RecordPairNetwork.

        super().__init__()
        self.left = left
        self.right = right
        self.matches = matches

        self.mismatch_share = mismatch_share
        self.shuffle = shuffle

        self.full_size = len(left) * len(right)
        self.used_size = self.full_size
        self.no_used_mismatches_per_match = None
        if self.matches is not None and self.mismatch_share < 1.0:
            no_matches = len(self.matches)
            no_total_mismatches = self.full_size - no_matches
            no_mismatches_per_match = max(no_total_mismatches // no_matches, 1)
            self.no_used_mismatches_per_match = max(
                int(self.mismatch_share * no_mismatches_per_match),
                1,
            )
            self.used_size = self.no_used_mismatches_per_match * no_matches + no_matches
        self.batch_size = int(min(batch_size, self.used_size))
        self.no_batches = int(np.floor(self.used_size / self.batch_size))
        self.last_batch_size = self.used_size % self.batch_size
        if self.last_batch_size > 0:
            self.no_batches += 1
        else:
            self.last_batch_size = self.batch_size

        self.similarity_map = similarity_map
        self.similarity_encoder = SimilarityEncoder(similarity_map)

        self.indices = self.__prepare_indices()

        self.on_epoch_end()

    def __prepare_indices(self) -> np.ndarray:
        if self.no_used_mismatches_per_match is None:
            return np.arange(self.used_size, dtype=np.int32)

        indices = np.array([], dtype=np.int32)
        for i, (li, ri) in enumerate(self.matches.itertuples(index=False)):
            indices = np.append(indices, li * len(self.right) + ri)
        assert len(indices) == len(self.matches)
        assert len(indices) <= self.used_size
        for li in range(len(self.left)):
            for ri in range(len(self.right)):
                if any(
                    (self.matches.iloc[:, 0] == li) & (self.matches.iloc[:, 1] == ri)
                ):
                    continue
                indices = np.append(indices, li * len(self.right) + ri)
                if len(indices) == self.used_size:
                    break
            if len(indices) == self.used_size:
                break
        assert len(indices) == self.used_size

        return indices

    def no_pairs(self) -> int:
        """Return the number of record pairs."""
        return self.used_size

    def no_matches(self) -> int:
        """Return the number of matches."""
        return len(self.matches)

    def no_mismatches(self) -> int:
        """Return the number of mismatches."""
        return self.used_size - self.no_matches()

    def __side_indices(
        self, indices: np.ndarray
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        lpos = np.array([k // len(self.right) for k in indices], dtype=np.int32)
        rpos = np.array([k % len(self.right) for k in indices], dtype=np.int32)
        return lpos, rpos

    def __labels(self, lpos: np.ndarray, rpos: np.ndarray) -> np.ndarray:
        if self.matches is None:
            return None
        labels = np.zeros((len(lpos)), dtype=np.float32)
        for i, lp in enumerate(lpos):
            labels[i] = any((self.matches.left == lp) & (self.matches.right == rpos[i]))
        return labels

    def __len__(self) -> int:
        """Return the number of batches per epoch."""
        return self.no_batches

    def __getitem__(
        self, index: int
    ) -> typing.Union[dict, typing.Tuple[dict, np.ndarray]]:
        """Get the batch at the given index."""
        begin = index * self.batch_size
        if index < self.no_batches - 1:
            end = begin + self.batch_size
        elif index == self.no_batches - 1:
            end = begin + self.last_batch_size
        else:
            raise IndexError("Invalid batch index")
        indices = self.indices[begin:end]
        lpos, rpos = self.__side_indices(indices)

        features = self.similarity_encoder(self.left.iloc[lpos], self.right.iloc[rpos])
        features = {
            key: features[i]
            for i, key in enumerate(self.similarity_map.association_names())
        }

        if self.matches is None:
            return features

        labels = self.__labels(lpos, rpos)
        return features, labels

    def on_epoch_end(self) -> None:
        """Maybe shuffle indices at the end of each epoch."""
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __str__(self) -> str:
        """Return a string representation of the data generator."""
        items = "\n  ".join(
            [
                self.similarity_map.__str__().replace("\n", "\n  "),
                f"No Batches[{self.no_batches}]",
                f"Batch size[{self.batch_size}]",
                f"No Pairs[{self.no_pairs()}]",
                f"No Matches[{self.no_matches()}]",
                f"No Mismatches[{self.no_mismatches()}]",
            ]
        )
        return f"{self.__class__.__name__}[\n  {items}]"



"""
Matching models module.

This module contains functionality for instantiating, training, and evaluating deep
 learning and neural-symbolic matching models
"""

from neer_match.axiom_generator import AxiomGenerator
from neer_match.data_generator import DataGenerator
from neer_match.record_pair_network import RecordPairNetwork
from neer_match.similarity_map import SimilarityMap
import ltn
import pandas as pd
import tensorflow as tf
import typing


def _suggest(
    model: typing.Union["DLMatchingModel", "NSMatchingModel"],
    left: pd.DataFrame,
    right: pd.DataFrame,
    count: int,
    batch_size: int = 32,
    **kwargs,
) -> pd.DataFrame:
    generator = DataGenerator(
        model.record_pair_network.similarity_map,
        left,
        right,
        mismatch_share=1.0,
        batch_size=batch_size,
        shuffle=False,
    )
    predictions = model.predict_from_generator(generator, **kwargs)[
        : len(left) * len(right)
    ]
    sides = generator._DataGenerator__side_indices(generator.indices)
    features = pd.DataFrame({"left": sides[0], "right": sides[1]})
    suggestions = features.assign(prediction=predictions)
    where = (
        suggestions.groupby((features.index / right.shape[0]).astype(int))["prediction"]
        .nlargest(count)
        .index.get_level_values(1)
    )
    return suggestions.iloc[where]


class DLMatchingModel(tf.keras.Model):
    """A deep learning matching model class.

    Inherits :class:`tensorflow.keras.Model` and automates deep-learning-based entity
    matching using the similarity map supplied by the user.

    Attributes:
        record_pair_network (RecordPairNetwork): The record pair network.
    """

    def __init__(
        self,
        similarity_map: SimilarityMap,
        initial_feature_width_scales: typing.Union[int, typing.List[int]] = 10,
        feature_depths: typing.Union[int, typing.List[int]] = 2,
        initial_record_width_scale: int = 10,
        record_depth: int = 4,
        **kwargs,
    ) -> None:
        """Initialize a deep learning matching model.

        Generate a record pair network from the passed similarity map. The input
        arguments are passed to the record pair network (see
        :class:`.RecordPairNetwork`).

        Args:
            similarity_map: A similarity map object.
            initial_feature_width_scales: The initial width scales of the feature
                networks.
            feature_depths: The depths of the feature networks.
            initial_record_width_scale: The initial width scale of the record network.
            record_depth: The depth of the record network.
            **kwargs: Additional keyword arguments passed to parent class
                      (:class:`tensorflow.keras.Model`).
        """
        super().__init__(**kwargs)
        self.record_pair_network = RecordPairNetwork(
            similarity_map,
            initial_feature_width_scales=initial_feature_width_scales,
            feature_depths=feature_depths,
            initial_record_width_scale=initial_record_width_scale,
            record_depth=record_depth,
        )

    def build(self, input_shapes: typing.List[tf.TensorShape]) -> None:
        """Build the model."""
        super().build(input_shapes)
        self.record_pair_network.build(input_shapes)

    def call(self, inputs: typing.Dict[str, tf.Tensor]) -> tf.Tensor:
        """Call the model on inputs."""
        return self.record_pair_network(inputs)

    def fit(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        matches: pd.DataFrame,
        batch_size: int = 16,
        mismatch_share: float = 0.1,
        shuffle: bool = True,
        **kwargs,
    ) -> None:
        """Fit the model.

        Construct a data generator from the input data frames using the
        similarity map with which the model was initialized and fit the model.
        The model is trained by calling the :func:`tensorflow.keras.Model.fit` method.

        Args:
            left: The left data frame.
            right: The right data frame.
            matches: The matches data frame.
            batch_size: Batch size.
            mismatch_share: Mismatch share.
            shuffle: Shuffle flag.
            **kwargs: Additional keyword arguments passed to parent class
                      (:func:`tensorflow.keras.Model.fit`).
        """
        generator = DataGenerator(
            self.record_pair_network.similarity_map,
            left,
            right,
            matches,
            batch_size=batch_size,
            mismatch_share=mismatch_share,
            shuffle=shuffle,
        )

        return super().fit(generator, **kwargs)

    def evaluate(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        matches: pd.DataFrame,
        **kwargs,
    ) -> dict:
        """Evaluate the model.

        Construct a data generator from the input data frames using the
        similarity map with which the model was initialized and evaluate the model.
        The model is evaluated by calling the :func:`tensorflow.keras.Model.evaluate`

        Args:
            left: The left data frame.
            right: The right data frame.
            matches: The matches data frame.
            **kwargs: Additional keyword arguments passed to parent class
                      (:func:`tensorflow.keras.Model.evaluate`).
        """
        generator = DataGenerator(
            self.record_pair_network.similarity_map,
            left,
            right,
            matches,
            mismatch_share=1.0,
            shuffle=False,
        )
        return super().evaluate(generator, **kwargs)

    def predict_from_generator(self, generator: DataGenerator, **kwargs) -> tf.Tensor:
        """Generate model predictions from a generator.

        Args:
            generator: The data generator.
            **kwargs: Additional keyword arguments passed to parent class
                      (:func:`tensorflow.keras.Model.predict`).
        """
        return super().predict(generator, **kwargs)

    def predict(
        self, left: pd.DataFrame, right: pd.DataFrame, batch_size: int = 16, **kwargs
    ) -> tf.Tensor:
        """Generate model predictions.

        Construct a data generator from the input data frames using the
        similarity map with which the model was initialized and generate predictions.

        Args:
            left: The left data frame.
            right: The right data frame.
            batch_size: Batch size.
            **kwargs: Additional keyword arguments passed to parent class
                      (:func:`tensorflow.keras.Model.predict`).
        """
        generator = DataGenerator(
            self.record_pair_network.similarity_map,
            left,
            right,
            batch_size=batch_size,
            mismatch_share=1.0,
            shuffle=False,
        )
        return self.predict_from_generator(generator, **kwargs)

    def suggest(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        count: int,
        batch_size: int = 16,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate model suggestions.

        Construct a data generator from the input data frames using the similarity map
        with which the model was initialized and generate suggestions.

        Args:
            left: The left data frame.
            right: The right data frame.
            count: The number of suggestions to generate.
            **kwargs: Additional keyword arguments passed to the suggest function.
        """
        return _suggest(self, left, right, count, batch_size=batch_size, **kwargs)

    @property
    def similarity_map(self) -> SimilarityMap:
        """Similarity Map of the Model."""
        return self.record_pair_network.similarity_map


class NSMatchingModel:
    """A neural-symbolic matching model class.

    Attributes:
        record_pair_network (RecordPairNetwork): The record pair network.
        bce (tf.keras.losses.Loss): The training loss function (binary cross-entropy, see
            :func:`tensorflow.keras.losses.BinaryCrossentropy`).
        optimizer (tensorflow.keras.optimizers.Optimizer): The optimizer used for
            training.
    """

    def __init__(
        self,
        similarity_map: SimilarityMap,
        initial_feature_width_scales: typing.Union[int, typing.List[int]] = 10,
        feature_depths: typing.Union[int, typing.List[int]] = 2,
        initial_record_width_scale: int = 10,
        record_depth: int = 4,
    ) -> None:
        """Initialize a neural-symbolic matching learning matching model.

        Generate a record pair network from the passed similarity map. The input
        arguments are passed to the record pair network (see
        :class:`.RecordPairNetwork`).

        The class uses a custom training loop with neural-symbolic (or hybrid) loss
        function. It does not inherit from :class:`tensorflow.keras.Model`, but to
        provide a consistent interface with the deep learning matching model, it
        implements the same methods.

        Args:
            similarity_map: A similarity map object.
            initial_feature_width_scales: The initial width scales of the feature
                networks.
            feature_depths: The depths of the feature networks.
            initial_record_width_scale: The initial width scale of the record network.
            record_depth: The depth of the record network.
        """
        self.record_pair_network = RecordPairNetwork(
            similarity_map,
            initial_feature_width_scales=initial_feature_width_scales,
            feature_depths=feature_depths,
            initial_record_width_scale=initial_record_width_scale,
            record_depth=record_depth,
        )

    def compile(
        self,
        optimizer: tf.keras.optimizers.Optimizer = tf.keras.optimizers.Adam(
            learning_rate=1e-4
        ),
    ) -> None:
        """Compile the model.

        Args:
            optimizer: The optimizer used for training.
        """
        self.bce = tf.keras.losses.BinaryCrossentropy()
        self.optimizer = optimizer
        features = [
            tf.random.normal((1, size), dtype=tf.float32)
            for size in self.record_pair_network.similarity_map.association_sizes()
        ]
        self.record_pair_network(features)

    def _make_axioms(self, data_generator: DataGenerator) -> typing.Callable:
        axiom_generator = AxiomGenerator(data_generator)
        field_predicates = [
            ltn.Predicate(f) for f in self.record_pair_network.field_networks
        ]
        record_predicate = ltn.Predicate(self.record_pair_network)
        is_positive = ltn.Predicate.Lambda(lambda x: x > 0.0)

        @tf.function
        def axioms(features: dict[str, tf.Tensor], labels: tf.Tensor) -> tf.Tensor:
            propositions = []

            y = ltn.Variable("y", labels)
            x = [
                ltn.Variable(f"x{i}", features[key])
                for i, key in enumerate(features.keys())
            ]

            stmt = axiom_generator.ForAll(
                [x[0], y], field_predicates[0](x[0]), mask=is_positive(y)
            )
            for i, F in enumerate(field_predicates[1:]):
                stmt = axiom_generator.Or(
                    stmt,
                    axiom_generator.ForAll(
                        [x[i + 1], y], F(x[i + 1]), mask=is_positive(y)
                    ),
                )
            propositions.append(stmt)
            stmt = axiom_generator.ForAll(
                [*x, y], record_predicate(x), mask=is_positive(y)
            )
            propositions.append(stmt)

            stmt = axiom_generator.ForAll(
                [x[0], y],
                axiom_generator.Not(field_predicates[0](x[0])),
                mask=axiom_generator.Not(is_positive(y)),
            )
            for i, F in enumerate(field_predicates[1:]):
                stmt = axiom_generator.Or(
                    stmt,
                    axiom_generator.ForAll(
                        [x[i + 1], y],
                        axiom_generator.Not(F(x[i + 1])),
                        mask=axiom_generator.Not(is_positive(y)),
                    ),
                )
            propositions.append(stmt)
            stmt = axiom_generator.ForAll(
                [*x, y],
                axiom_generator.Not(record_predicate(x)),
                mask=axiom_generator.Not(is_positive(y)),
            )
            propositions.append(stmt)

            kb = axiom_generator.FormAgg(propositions)
            sat = kb.tensor
            return sat

        return axioms

    def __for_epoch(
        self,
        data_generator: DataGenerator,
        loss_clb: typing.Callable,
        trainable_variables: typing.List[tf.Variable],
        verbose: int = 1,
    ) -> dict:
        no_batches = len(data_generator)
        pb_size = 60
        logs = {
            "no_batches": no_batches,
            "TP": 0,
            "FP": 0,
            "TN": 0,
            "FN": 0,
            "BCE": 0,
            "Sat": 0,
            "Loss": 0,
        }

        for i, (features, labels) in enumerate(data_generator):
            if verbose > 0:
                pb_step = int((i + 1) / no_batches * pb_size)
                pb = "=" * pb_step + "." * (pb_size - pb_step)
                print(f"\r[{pb}] {i + 1}/{no_batches}", end="", flush=True)
            with tf.GradientTape() as tape:
                batch_loss, batch_logs = loss_clb(features, labels)
            grads = tape.gradient(batch_loss, trainable_variables)
            self.optimizer.apply_gradients(zip(grads, trainable_variables))

            preds = batch_logs["Predicted"]
            preds = tf.reshape(preds, preds.shape[0])
            logs["TP"] += tf.reduce_sum(tf.round(preds) * labels)
            logs["FP"] += tf.reduce_sum(tf.round(preds) * (1.0 - labels))
            logs["TN"] += tf.reduce_sum((1.0 - tf.round(preds)) * (1.0 - labels))
            logs["FN"] += tf.reduce_sum((1.0 - tf.round(preds)) * labels)
            logs["BCE"] += batch_logs["BCE"]
            logs["Sat"] += batch_logs["Sat"]
            if "ASat" in batch_logs:
                if "ASat" not in logs:
                    logs["ASat"] = batch_logs["ASat"]
                else:
                    logs["ASat"] += batch_logs["ASat"]
            logs["Loss"] += batch_loss

        logs["Sat"] /= no_batches
        if "ASat" in logs:
            logs["ASat"] /= no_batches

        if verbose > 0:
            print("\r", end="", flush=True)

        return logs

    def __make_loss(
        self, axioms: typing.Callable, satisfiability_weight: float
    ) -> typing.Callable:
        @tf.function
        def loss_clb(
            features: dict[str, tf.Tensor], labels: tf.Tensor
        ) -> typing.Tuple[tf.Tensor, dict]:
            preds = self.record_pair_network(features)
            bce = self.bce(labels, preds)
            sat = axioms(features, labels)
            loss = (1.0 - satisfiability_weight) * bce + satisfiability_weight * (
                1.0 - sat
            )
            logs = {"BCE": bce, "Sat": sat, "Predicted": preds}
            return loss, logs

        return loss_clb

    def _training_loop_log_header(self) -> str:
        headers = ["Epoch", "BCE", "Recall", "Precision", "F1", "Sat"]
        return "| " + " | ".join([f"{x:<10}" for x in headers]) + " |"

    def _training_loop_log_row(self, epoch: int, logs: dict) -> str:
        recall = logs["TP"] / (logs["TP"] + logs["FN"])
        precision = logs["TP"] / (logs["TP"] + logs["FP"])
        f1 = 2.0 * precision * recall / (precision + recall)
        values = [logs["BCE"].numpy(), recall, precision, f1, logs["Sat"].numpy()]
        row = f"| {epoch:<10} | " + " | ".join([f"{x:<10.4f}" for x in values]) + " |"
        return row

    def _training_loop_log_end(self, epoch: int, logs: dict) -> str:
        return (
            f"Training finished at Epoch {epoch} with "
            f"DL loss {logs['BCE'].numpy():.4f} and "
            f"Sat {logs['Sat'].numpy():.4f}"
        )

    def _training_loop(
        self,
        data_generator: DataGenerator,
        loss_clb: typing.Callable,
        trainable_variables: typing.List[tf.Variable],
        epochs: int,
        verbose: int,
        log_mod_n: int,
    ) -> None:
        if verbose > 0:
            print(self._training_loop_log_header())

        for epoch in range(epochs):
            logs = self.__for_epoch(
                data_generator,
                loss_clb,
                trainable_variables,
                verbose=verbose - 1,
            )
            if verbose > 0 and epoch % log_mod_n == 0:
                print(self._training_loop_log_row(epoch, logs))
        if verbose > 0:
            print(self._training_loop_log_end(epoch, logs))

    def fit(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        matches: pd.DataFrame,
        epochs: int,
        mismatch_share: float = 0.1,
        satisfiability_weight: float = 1.0,
        verbose: int = 1,
        log_mod_n: int = 1,
        **kwargs,
    ) -> None:
        """Fit the model.

        Construct a data generator from the input data frames using the similarity map
        with which the model was initialized and fit the model.

        The model is trained using a custom training loop. The loss can either be purely
        defined using fuzzy logic axioms (default case with satisfiability weight 1.0)
        or as a weighted sum of binary cross-entropy and satisfiability loss (by setting
        the satisfiability weight to a value between 0 and 1).

        Args:
            left: The left data frame.
            right: The right data frame.
            matches: The matches data frame.
            epochs: The number of epochs to train.
            mismatch_share: The mismatch share.
            satisfiability_weight: The weight of the satisfiability loss.
            verbose: The verbosity level.
            log_mod_n: The log modulo.
            **kwargs: Additional keyword arguments passed to the data generator.
        """
        if not isinstance(epochs, int) or epochs < 1:
            raise ValueError("Epochs must be an integer greater than 0")
        if satisfiability_weight < 0.0 or satisfiability_weight > 1.0:
            raise ValueError("Satisfiability weight must be between 0 and 1")
        if not isinstance(verbose, int):
            raise ValueError("Verbose must be an integer")
        if not isinstance(log_mod_n, int) or log_mod_n < 1:
            raise ValueError("Log mod n must be an integer greater than 0")
        # The remaining arguments are validated in the DataGenerator

        data_generator = DataGenerator(
            self.record_pair_network.similarity_map, 
            left, 
            right, 
            matches, 
            mismatch_share=mismatch_share,
            **kwargs
        )

        axioms = self._make_axioms(data_generator)
        loss_clb = self.__make_loss(axioms, satisfiability_weight)

        trainable_variables = self.record_pair_network.trainable_variables
        self._training_loop(
            data_generator, loss_clb, trainable_variables, epochs, verbose, log_mod_n
        )

    def evaluate(
        self,
        left: pd.DataFrame,
        right: pd.DataFrame,
        matches: pd.DataFrame,
        batch_size: int = 16,
        mismatch_share: float = 1.0,
        satisfiability_weight: float = 1.0,
    ) -> dict:
        """Evaluate the model.

        Construct a data generator from the input data frames using the similarity map
        with which the model was initialized and evaluate the model. It returns a
        dictionary with evaluation metrics.

        Args:
            left: The left data frame.
            right: The right data frame.
            matches: The matches data frame.
            batch_size: Batch size.
            mismatch_share: The mismatch share.
            satisfiability_weight: The weight of the satisfiability loss.
        """
        data_generator = DataGenerator(
            self.record_pair_network.similarity_map,
            left,
            right,
            matches,
            mismatch_share=mismatch_share,
            batch_size=batch_size,
            shuffle=False,
        )

        axioms = self._make_axioms(data_generator)
        loss_clb = self.__make_loss(axioms, satisfiability_weight)

        trainable_variables = self.record_pair_network.trainable_variables
        logs = self.__for_epoch(
            data_generator,
            loss_clb,
            trainable_variables,
            verbose=1,
        )

        tp = logs["TP"]
        fp = logs["FP"]
        tn = logs["TN"]
        fn = logs["FN"]
        logs["Accuracy"] = (tp + tn) / (tp + tn + fp + fn)
        logs["Recall"] = tp / (tp + fn)
        logs["Precision"] = tp / (tp + fp)
        logs["F1"] = (
            2.0
            * logs["Precision"]
            * logs["Recall"]
            / (logs["Precision"] + logs["Recall"])
        )
        return {
            key: value.numpy() for key, value in logs.items() if key != "no_batches"
        }

    def predict_from_generator(self, generator: DataGenerator) -> tf.Tensor:
        """Generate model predictions from a generator.

        Args:
            generator: The data generator.
        """
        preds = self.record_pair_network(generator[0])
        for i, features in enumerate(generator):
            if i == 0:
                continue
            preds = tf.concat([preds, self.record_pair_network(features)], axis=0)
        return preds.numpy()

    def predict(
        self, left: pd.DataFrame, right: pd.DataFrame, batch_size: int = 16
    ) -> tf.Tensor:
        """Generate model predictions.

        Construct a data generator from the input data frames using the
        similarity map with which the model was initialized and generate predictions.

        Args:
            left: The left data frame.
            right: The right data frame.
            batch_size: Batch size.
        """
        generator = DataGenerator(
            self.record_pair_network.similarity_map,
            left,
            right,
            mismatch_share=1.0,
            batch_size=batch_size,
            shuffle=False,
        )
        return self.predict_from_generator(generator)

    def suggest(
        self, left: pd.DataFrame, right: pd.DataFrame, count: int, batch_size=16
    ) -> pd.DataFrame:
        """Generate model suggestions.

        Construct a data generator from the input data frames using the similarity map
        with which the model was initialized and generate suggestions.

        Args:
            left: The left data frame.
            right: The right data frame.
            count: The number of suggestions to generate.
            batch_size: Batch size.
        """
        return _suggest(self, left, right, count, batch_size=batch_size)

    @property
    def similarity_map(self) -> SimilarityMap:
        """Similarity Map of the Model."""
        return self.record_pair_network.similarity_map


def _matching_model_or_raise(
    model: typing.Union[DLMatchingModel, NSMatchingModel]
) -> None:
    if not isinstance(model, (DLMatchingModel, NSMatchingModel)):
        raise ValueError(
            "The model argument must be an instance of DLMatchingModel "
            "or NSMatchingModel"
        )



"""
Similarity mappings module.

The module provides functionality to store and manage a similarity mappings between
records of two datasets.
"""

from rapidfuzz import distance, fuzz
import numpy
import typing


def discrete(
    x: typing.Union[int, str, typing.AnyStr], y: typing.Union[int, str, typing.AnyStr]
) -> float:
    """Discrete similarity function."""
    return 1.0 if x == y else 0.0


def euclidean(x: typing.Union[float, int], y: typing.Union[float, int]) -> float:
    """Euclidean similarity function."""
    return 1.0 / (1.0 + abs(x - y))


def gaussian(x: typing.Union[float, int], y: typing.Union[float, int]) -> float:
    """Gaussian similarity function."""
    return numpy.exp(-((x - y) ** 2) / 2.0)


def available_similarities() -> typing.Dict[str, typing.Callable]:
    """Return the list of available similarities."""
    return {
        "basic_ratio": fuzz.ratio,
        "damerau_levenshtein": distance.DamerauLevenshtein.normalized_similarity,
        "discrete": discrete,
        "euclidean": euclidean,
        "gaussian": gaussian,
        "hamming": distance.Hamming.normalized_similarity,
        "indel": distance.Indel.normalized_similarity,
        "jaro": distance.Jaro.normalized_similarity,
        "jaro_winkler": distance.JaroWinkler.normalized_similarity,
        "lcsseq": distance.LCSseq.normalized_similarity,
        "levenshtein": distance.Levenshtein.normalized_similarity,
        "osa": distance.OSA.normalized_similarity,
        "partial_ratio": fuzz.partial_ratio,
        "partial_ratio_alignment": fuzz.partial_ratio_alignment,
        "partial_token_ratio": fuzz.partial_token_ratio,
        "partial_token_set_ratio": fuzz.partial_token_set_ratio,
        "partial_token_sort_ratio": fuzz.partial_token_sort_ratio,
        "postfix": distance.Postfix.normalized_similarity,
        "prefix": distance.Prefix.normalized_similarity,
        "token_ratio": fuzz.token_ratio,
        "token_set_ratio": fuzz.token_set_ratio,
        "token_sort_ratio": fuzz.token_sort_ratio,
    }


class SimilarityMap:
    """
    Similarity map class.

    The class stores a collection of associations between the records of two datasets.

    Attributes:
        instructions (dict): The similarity map instructions.
        lcols (list[str]): The left columns.
        rcols (list[str]): The right columns.
        sims (list[str]): The similarity functions.
    """

    def __init__(self, instructions):
        """Initialize a similarity map object.

        Args:
            instructions: The similarity map instructions.
        """
        if not isinstance(instructions, dict):
            raise ValueError("Input instructions must be a dictionary.")
        for key, value in instructions.items():
            if not isinstance(key, str):
                raise ValueError("Association key must be a string.")
            if not isinstance(value, list):
                raise ValueError(
                    "Association values must be a list. "
                    f"Instead got type {type(value)}."
                )
        self.instructions = instructions
        self.lcols = []
        self.rcols = []
        self.sims = []
        for association, similarities in self.instructions.items():
            parts = association.split("~")
            assert len(parts) == 1 or len(parts) == 2
            lcol = parts[0].strip()
            rcol = parts[1].strip() if len(parts) == 2 else lcol
            for similarity in similarities:
                self.lcols.append(lcol)
                self.rcols.append(rcol)
                self.sims.append(similarity)

    def no_associations(self) -> int:
        """Return the number of associations of the map."""
        return len(self.instructions)

    def association_sizes(self) -> typing.List[int]:
        """Return then number of similarities used by each association."""
        return [len(instruction) for instruction in self.instructions.values()]

    def association_offsets(self) -> typing.List[int]:
        """Return association offsets.

        Return the starting column offset of each association in the
        similarity matrix
        """
        return numpy.cumsum([0] + self.association_sizes()).tolist()[0:-1]

    def association_names(self) -> typing.List[str]:
        """Return a unique name for each association in the similarity map."""
        names = []
        for key in self.instructions.keys():
            parts = [p.strip() for p in key.split("~")]
            names.append(parts[0] if len(parts) == 1 else f"{parts[0]}_{parts[1]}")
        return names

    def keys(self) -> typing.List[str]:
        """Return a unique key for each similarity map entry.

        Combine association with similarity names and return them.
        """
        keys = []
        for key, value in self.instructions.items():
            parts = [p.strip() for p in key.split("~")]
            name = parts[0] if len(parts) == 1 else f"{parts[0]}_{parts[1]}"
            for similarity in value:
                keys.append(f"{name}_{similarity}")
        return keys

    def __iter__(self) -> typing.Iterator:
        """Iterate over the similarity map."""
        return zip(self.lcols, self.rcols, self.sims)

    def __len__(self) -> int:
        """Return the number of items in the similarity map."""
        return len(self.lcols)

    def __getitem__(self, index) -> typing.Tuple[str, str, str]:
        """Return the item at the given index."""
        return self.lcols[index], self.rcols[index], self.sims[index]

    def __str__(self) -> str:
        """Return a string representation of the similarity map."""
        items = "\n  ".join([str(item) for item in self])
        return f"{self.__class__.__name__}[\n  {items}]"



"""
Reasoning module.

This module provides reasoning functionality for the entity matching tasks using logic
tensor networks.
"""

from neer_match.axiom_generator import AxiomGenerator
from neer_match.data_generator import DataGenerator
from neer_match.matching_model import NSMatchingModel
import tensorflow as tf
import ltn


class RefutationModel(NSMatchingModel):
    """A neural-symbolic refutation model for entity matching tasks.

    Inherits :class:`neer_match.matching_model.NSMatchingModel` and provides additional
    functionality for refutation logic. The built-in refutation logic allows one to
    refute the significance of one or more similarities of a  conjectured association
    in detecting entity matches.
    """

    def __make_claims(self, axiom_generator, refutation):
        instructions = axiom_generator.data_generator.similarity_map.instructions
        rassoc = list(refutation.keys())[0]
        rsim = list(refutation.values())[0]
        if len(rsim) == 0:
            sindices = range(len(instructions[rassoc]))
        else:
            sindices = [instructions[rassoc].index(sim) for sim in rsim]

        id_predicate = ltn.Predicate.Lambda(lambda x: x)

        @tf.function
        def claims(features, labels):
            propositions = []

            y = ltn.Variable("y", self.record_pair_network(features))
            x = [
                ltn.Variable(f"x{sindex}", features[rassoc][:, sindex])
                for sindex in sindices
            ]
            stmt = axiom_generator.ForAll(
                [x[0], y], axiom_generator.Implies(id_predicate(y), id_predicate(x[0]))
            )
            for v in x[1:]:
                stmt = axiom_generator.Or(
                    stmt,
                    axiom_generator.ForAll(
                        [v, y],
                        axiom_generator.Implies(id_predicate(y), id_predicate(v)),
                    ),
                )
            propositions.append(stmt)

            kb = axiom_generator.FormAgg(propositions)
            sat = kb.tensor
            return sat

        return claims

    def __make_loss(self, q, claims, satisfiability_weight, axioms, beta, alpha):
        @tf.function
        def loss_clb(features, labels):
            preds = self.record_pair_network(features)
            bce = self.bce(labels, preds)
            asat = axioms(features, labels)
            csat = claims(features, labels)
            loss = (1.0 - satisfiability_weight) * (
                1.0 - bce
            ) + satisfiability_weight * asat
            loss = csat + tf.keras.activations.elu(beta * (q - loss), alpha=alpha)
            logs = {"BCE": bce, "ASat": asat, "Sat": csat, "Predicted": preds}
            return loss, logs

        return loss_clb

    def _training_loop_log_header(self):
        headers = ["Epoch", "BCE", "Recall", "Precision", "F1", "CSat", "ASat"]
        return "| " + " | ".join([f"{x:<10}" for x in headers]) + " |"

    def _training_loop_log_row(self, epoch, logs):
        recall = logs["TP"] / (logs["TP"] + logs["FN"])
        precision = logs["TP"] / (logs["TP"] + logs["FP"])
        f1 = 2.0 * precision * recall / (precision + recall)
        values = [
            logs["BCE"].numpy(),
            recall,
            precision,
            f1,
            logs["Sat"].numpy(),
            logs["ASat"].numpy(),
        ]
        row = f"| {epoch:<10} | " + " | ".join([f"{x:<10.4f}" for x in values]) + " |"
        return row

    def __validate_refutation_arg(self, refutation):
        if isinstance(refutation, str):
            if refutation not in self.similarity_map.instructions.keys():
                raise ValueError(f"Association {refutation} is not in similarity map.")
            refutation = {refutation: self.similarity_map.instructions[refutation]}
        elif isinstance(refutation, dict):
            if len(refutation.keys()) != 1:
                raise ValueError(
                    "Refutation model can be initialized with only one refutation "
                    f"association. Instead got {len(refutation.keys())}."
                )
            key = list(refutation.keys())[0]
            if refutation[key] is None:
                refutation = {key: self.similarity_map.instructions[key]}
            elif isinstance(refutation[key], str):
                if refutation[key] not in self.similarity_map.instructions[key]:
                    raise ValueError(
                        f"Similarity {refutation[key]} is not in association {key}."
                    )
                refutation[key] = [refutation[key]]
            elif isinstance(refutation[key], list):
                if len(refutation[key]) == 0:
                    refutation = {key: self.similarity_map.instructions[key]}
                for similarity in refutation[key]:
                    if similarity not in self.similarity_map.instructions[key]:
                        raise ValueError(
                            f"Similarity {similarity} is not in association {key}."
                        )
            else:
                raise ValueError(
                    "Refutation similarities can be none or a list of similarities."
                    f"Instead got type {type(refutation[key])}."
                )
        else:
            raise ValueError(
                "Refutation must be a string or a dictionary. "
                f"Instead got type {type(refutation)}."
            )
        return refutation

    def fit(
        self,
        left,
        right,
        matches,
        epochs,
        refutation,
        penalty_threshold=0.95,
        penalty_scale=1.0,
        penalty_decay=0.1,
        satisfiability_weight=1.0,
        verbose=1,
        log_mod_n=1,
        **kwargs,
    ):
        """Fit the refutation model.

        Construct a data generator and an axiom generator from the input data and
        use the model's similarity map to fit the model while trying to refute the
        refutation claim.

        In the default case of satisfiability weight equal to 1, the function
        minimizes the satisfiability of the refutation claim
        while penalizing the satisfiability of the matching axioms below the
        penalty threshold. If the satisfiability weight is less than 1, the model is
        trained to optimize the satisfiability of the refutation claim, while penalizing
        a weighted sum of the satisfiability of the matching axioms and the binary cross
        entropy loss for values below the penalty threshold.

        The penalty threshold sets tolerance for the matching axioms (and/or the
        binary cross entropy loss) below which the penalty is applied. The penalty scale
        sets the linear scale of the penalty when the threshold is not crossed. The
        penalty decay sets the exponential decay of the penalty when the threshold is
        crossed. The linear and exponential parts are combined using the
        :func:`tensorflow.keras.activations.elu` function.

        Args:
            left: The left dataset.
            right: The right dataset.
            matches: The matches dataset.
            epochs: The number of epochs to train the model.
            refutation: The refutation claim. It can be a string or a dictionary.
                If it is a string, it is assumed to be an association name. If it is
                a dictionary, the keys are association names and the values are
                similarity names. If the value is None, all similarities in the
                association are used.
            penalty_threshold: The penalty threshold.
            penalty_scale: The non-satisfiability scale.
            penalty_decay: The non-satisfiability decay.
            satisfiability_weight: The satisfiability weight.
            verbose: The verbosity level.
            log_mod_n: The logging frequency.
            **kwargs: Additional arguments to the data generator.

        """
        refutation = self.__validate_refutation_arg(refutation)
        if penalty_threshold < 0.0 or penalty_threshold > 1.0:
            raise ValueError("Satisfiability threshold must be between 0 and 1")
        if penalty_scale < 0.0:
            raise ValueError("Non-satisfiability scale must be positive")
        if penalty_decay < 0.0:
            raise ValueError("Satisfiability scale must be positive")
        if satisfiability_weight < 0.0 or satisfiability_weight > 1.0:
            raise ValueError("Satisfiability weight must be between 0 and 1")
        # The remaining arguments are checked by the parent class' fit method

        data_generator = DataGenerator(
            self.record_pair_network.similarity_map, left, right, matches, **kwargs
        )
        axiom_generator = AxiomGenerator(data_generator)

        axioms = self._make_axioms(axiom_generator)
        claims = self.__make_claims(axiom_generator, refutation)
        loss_clb = self.__make_loss(
            penalty_threshold,
            claims,
            satisfiability_weight,
            axioms,
            penalty_scale,
            penalty_decay,
        )

        trainable_variables = self.record_pair_network.trainable_variables
        super()._training_loop(
            data_generator, loss_clb, trainable_variables, epochs, verbose, log_mod_n
        )



"""
Entity matching axiom generator module.

This module provides an axiom generation functionality for entity matching tasks.
"""

from neer_match.data_generator import DataGenerator
import ltn
import numpy as np
import tensorflow as tf
import typing


class AxiomGenerator:
    """Axiom generator class.

    The class contains the definitions for the logical connectives, functions,
    quantifiers, and aggregators used in the axiom generation process.

    Attributes:
        data_generator (DataGenerator): A data generator object.
        Sim (ltn.Predicate.Lambda): A similarity predicate.
        Not (ltn.Wrapper_Connective): A negation connective.
        And (ltn.Wrapper_Connective): A conjunction connective.
        Or (ltn.Wrapper_Connective): A disjunction connective.
        Implies (ltn.Wrapper_Connective): An implication connective.
        Equiv (ltn.Wrapper_Connective): An equivalence connective.
        ForAll (ltn.Wrapper_Quantifier): A universal quantifier.
        Exists (ltn.Wrapper_Quantifier): An existential quantifier.
        FormAgg (ltn.Wrapper_Formula_Aggregator): A formula aggregator.
    """

    def __init__(self, data_generator: DataGenerator) -> None:
        r"""Initialize an axiom generator object.

        By default, the axiom generator uses product t-norm and co t-norm for
        conjunction and disjunction, respectively. In addition, it uses Reichenbach
        implications. The quantifiers are based on the p-mean error and p-mean
        aggregators (see also :ltn-tut-2c:`\ `).

        Args:
            data_generator: A data generator object.
        """
        self.data_generator = data_generator

        self.Sim = ltn.Predicate.Lambda(
            lambda args: tf.exp(-1.0 * tf.norm(args[0] - args[1], axis=1, ord=np.inf))
        )
        self.Not = ltn.Wrapper_Connective(ltn.fuzzy_ops.Not_Std())
        self.And = ltn.Wrapper_Connective(ltn.fuzzy_ops.And_Prod())
        self.Or = ltn.Wrapper_Connective(ltn.fuzzy_ops.Or_ProbSum())
        self.Implies = ltn.Wrapper_Connective(ltn.fuzzy_ops.Implies_Reichenbach())
        self.Equiv = ltn.Wrapper_Connective(
            ltn.fuzzy_ops.Equiv(
                ltn.fuzzy_ops.And_Prod(), ltn.fuzzy_ops.Implies_Reichenbach()
            )
        )
        self.ForAll = ltn.Wrapper_Quantifier(
            ltn.fuzzy_ops.Aggreg_pMeanError(p=2), semantics="forall"
        )
        self.Exists = ltn.Wrapper_Quantifier(
            ltn.fuzzy_ops.Aggreg_pMean(p=2), semantics="exists"
        )
        self.FormAgg = ltn.Wrapper_Formula_Aggregator(
            ltn.fuzzy_ops.Aggreg_pMeanError(p=2)
        )

    def __select_field_constants(
        self, value: float
    ) -> typing.List[typing.List[ltn.Constant]]:
        examples = self.data_generator._DataGenerator__select_features(value)
        consts = [
            [ltn.Constant(point, trainable=False) for point in feature]
            for i, feature in enumerate(examples)
        ]
        return consts

    def field_matching_constants(self) -> typing.List[typing.List[ltn.Constant]]:
        """Return field matching constants.

        The outer list contains contains matching examples and the inner lists contain
        the constants of the features.
        """
        return self.__select_field_constants(1.0)

    def field_non_matching_constants(self) -> typing.List[typing.List[ltn.Constant]]:
        """Return field non matching constants.

        The outer list contains contains non-matching examples and the inner lists
        contain the constants of the features.
        """
        return self.__select_field_constants(0.0)
